const express = require("express");
const Game = require("./game.js");

const GamePhase = {
  JOIN:    1,
  REVEAL:  2,
  CAPTURE: 3,
  properties: {
    1: {name: "join"},
    2: {name: "reveal"},
    3: {name: "capture"}
  }
};

/**
 * Shuffles array in place.
 * @param {Array} a items An array containing the items.
 */
function shuffle(a) {
  if (a === undefined)
    return;

  var j, x, i;
  for (i = a.length - 1; i > 0; i--) {
    j = Math.floor(Math.random() * (i + 1));
    x = a[i];
    a[i] = a[j];
    a[j] = x;
  }
}

class Empires extends Game {
  constructor(id) {
    super(id);
    this._spectators = [];
    this.next_spectator_id = 1;
    this._players = [];
    this.next_player_id = 1;
    this.characters_to_reveal = null;
    this.phase = GamePhase.JOIN;
  }

  add_spectator(name, session_id) {
    /* Don't add another spectator that matches an existing session. */
    const existing = this._spectators.findIndex(
      spectator => spectator.session_id === session_id);
    if (existing >= 0)
      return existing.id;

    const new_spectator = {id: this.next_spectator_id,
                           name: name,
                           session_id: session_id
                          };
    this._spectators.push(new_spectator);
    this.next_spectator_id++;
    this.broadcast_event("spectator-join", JSON.stringify(new_spectator));

    return new_spectator.id;
  }

  remove_spectator(id) {
    const index = this._spectators.findIndex(spectator => spectator.id === id);
    this._spectators.splice(index, 1);

    this.broadcast_event("spectator-leave", `{"id": ${id}}`);
  }

  register_player(name, character) {
    const new_player = {id: this.next_player_id,
                       name: name,
                       character: character,
                       captures: [],
                       };
    this._players.push(new_player);
    this.next_player_id++;
    /* The syntax here is using an anonymous function to create a new
      object from new_player with just the subset of fields that we
      want. */
    const player_data = JSON.stringify((({id, name}) => ({id, name}))(new_player));
    this.broadcast_event("player-join", player_data);

    return new_player;
  }

  remove_player(id) {
    const index = this._players.findIndex(player => player.id === id);
    this._players.splice(index, 1);

    this.broadcast_event("player-leave", `{"id": ${id}}`);
  }

  reset() {
    this._players = [];
    this.characters_to_reveal = null;
    this.next_player_id = 1;

    this.change_phase(GamePhase.JOIN);

    this.broadcast_event("spectators", "{}");
    this.broadcast_event("players", "{}");
  }

  reveal_next() {
    /* Don't try to reveal anything if we aren't in the reveal phase. */
    if (this.phase != GamePhase.REVEAL) {
      clearInterval(this.reveal_interval);
      return;
    }

    if (this.reveal_index >= this.characters_to_reveal.length) {
      clearInterval(this.reveal_interval);
      this.broadcast_event("character-reveal", '{"character":""}');
      return;
    }
    const character = this.characters_to_reveal[this.reveal_index];
    this.reveal_index++;
    const character_data = JSON.stringify({"character":character});
    this.broadcast_event("character-reveal", character_data);
  }

  reveal() {
    this.change_phase(GamePhase.REVEAL);

    if (this.characters_to_reveal === null) {
      this.characters_to_reveal = [];
      this.characters_to_reveal = this._players.reduce((characters, player) => {
        characters.push(player.character);
        return characters;
      }, []);
      shuffle(this.characters_to_reveal);
    }

    this.reveal_index = 0;

    this.reveal_interval = setInterval(this.reveal_next.bind(this), 3000);
  }

  start() {
    this.change_phase(GamePhase.CAPTURE);
  }

  capture(captor_id, captee_id) {
    /* TODO: Fix to fail on already-captured players (or to move the
     * captured player from an old captor to a new—need to clarify in
     * the API specification which we want here. */
    let captor = this._players.find(player => player.id === captor_id);
    captor.captures.push(captee_id);

    this.broadcast_event("capture", `{"captor": ${captor_id}, "captee": ${captee_id}}`);
  }

  liberate(captee_id) {
    let captor = this._players.find(player => player.captures.includes(captee_id));
    captor.captures.splice(captor.captures.indexOf(captee_id), 1);
  }

  restart() {
    for (const player of this._players) {
      player.captures = [];
    }
  }

  get characters() {
    return this._players.map(player => player.character);
  }

  get empires() {
    return this._players.map(player => ({id: player.id, captures: player.captures}));
  }

  get spectators() {
    /* We return only "id" and "name" here (specifically not session_id!). */
    return this._spectators.map(spectator => ({
      id: spectator.id,
      name: spectator.name
    }));
  }

  /* The base class recently acquired Game.players which works like
   * Empires.spectators, (and meanwhile the Empires._players
   * functionality could perhaps be reworked into
   * Game.players[].team). Until we do that rework, lets use
   * .registered_players as the getter for the Empires-specific
   * ._players property to avoid mixing it up with the distinct
   * Game.players property. */
  get registered_players() {
    return this._players.map(player => ({id: player.id, name: player.name }));
  }

  game_phase_event_data(old_phase, new_phase) {
    var old_phase_name;
    if (old_phase)
      old_phase_name = GamePhase.properties[old_phase].name;
    else
      old_phase_name = "none";
    const new_phase_name = GamePhase.properties[new_phase].name;

    return `{"old_phase":"${old_phase_name}","new_phase":"${new_phase_name}"}`;
  }

  /* Inform clients about a phase change. */
  broadcast_phase_change() {
    const event_data = this.game_phase_event_data(this.old_phase, this.phase);
    this.broadcast_event("game-phase", event_data);
  }

  /* Change game phase and broadcast the change to all clients. */
  change_phase(phase) {
    /* Do nothing if there is no actual change happening. */
    if (phase === this.phase)
      return;

    this.old_phase = this.phase;
    this.phase = phase;

    this.broadcast_phase_change();
  }

  handle_events(request, response) {

    super.handle_events(request, response);

    /* Now that a client has connected, first we need to stream all of
     * the existing spectators and players (if any). */
    if (this._spectators.length > 0) {
      const spectators_json = JSON.stringify(this.spectators);
      const spectators_data = `event: spectators\ndata: ${spectators_json}\n\n`;
      response.write(spectators_data);
    }

    if (this._players.length > 0) {
      const players_json = JSON.stringify(this.registered_players);
      const players_data = `event: players\ndata: ${players_json}\n\n`;
      response.write(players_data);
    }

    /* And we need to inform the client of the current game phase.
     *
     * In fact, we need to cycle through each phase transition from the
     * beginning so the client can see each.
     */
    var old_phase = null;
    for (var phase = GamePhase.JOIN; phase <= this.phase; phase++) {
      var event_data = this.game_phase_event_data(old_phase, phase);
      response.write("event: game-phase\n" + "data: " + event_data + "\n\n");
      old_phase = phase;
    }
  }

}

Empires.router = express.Router();
const router = Empires.router;

router.post('/spectator', (request, response) => {
  const game = request.game;
  var name = request.session.nickname;

  /* If the request includes a name, that overrides the session nickname. */
  if (request.body.name)
    name = request.body.name;

  const id = game.add_spectator(name, request.session.id);
  response.send(JSON.stringify(id));
});

router.delete('/spectator/:id', (request, response) => {
  const game = request.game;
  game.remove_spectator(parseInt(request.params.id));
  response.send();
});

router.post('/register', (request, response) => {
  const game = request.game;
  var name = request.session.nickname;;

  /* If the request includes a name, that overrides the session nickname. */
  if (request.body.name)
    name = request.body.name;

  const player = game.register_player(name, request.body.character);
  response.send(JSON.stringify(player.id));
});

router.post('/deregister/:id', (request, response) => {
  const game = request.game;
  game.remove_player(parseInt(request.params.id));
  response.send();
});

router.post('/reveal', (request, response) => {
  const game = request.game;
  game.reveal();
  response.send();
});

router.post('/start', (request, response) => {
  const game = request.game;
  game.start();
  response.send();
});

router.post('/reset', (request, response) => {
  const game = request.game;
  game.reset();
  response.send();
});

router.post('/capture/:captor/:captee', (request, response) => {
  const game = request.game;
  game.capture(parseInt(request.params.captor), parseInt(request.params.captee));
  response.send();
});

router.post('/liberate/:id', (request, response) => {
  const game = request.game;
  game.liberate(parseInt(request.params.id));
  response.send();
});

router.post('/restart', (request, response) => {
  const game = request.game;
  game.restart(parseInt(request.params.id));
    response.send();
});

router.get('/characters', (request, response) => {
  const game = request.game;
  response.send(game.characters);
});

router.get('/empires', (request, response) => {
  const game = request.game;
  response.send(game.empires);
});

router.get('/spectators', (request, response) => {
  const game = request.game;
  response.send(game.spectators);
});

router.get('/players', (request, response) => {
  const game = request.game;
  response.send(game.registered_players);
});

Empires.meta = {
  name: "Empires",
  identifier: "empires"
};

exports.Game = Empires;
